from typing import List, Any, Dict, Tuple
import uuid
import os
import shutil
import pandas as pd
from sqlalchemy import Column
from sqlalchemy.exc import InvalidRequestError
from enums.CatalogConfigurationEnum import CatalogConfigurationEnum
from enums.FileTypeEnum import FileTypeEnum
from enums.DataTypeEnum import DataTypeEnum
from enums.DataTypeOrmEnum import DataTypeOrmEnum
from components.Model.InsumoModel import InsumoModel
from enums.CommentsScriptEnum import CommentsScriptEnum
from components.Timezone import datetime_by_tzone

import logging

logger = logging.getLogger()


def get_type_file(key: str) -> FileTypeEnum:
    result = FileTypeEnum.EXCEL
    try:
        file_type_sufix = key.rfind(".")
        file_type = key[file_type_sufix+1:]
        result = FileTypeEnum(file_type)
    except Exception as e:
        logger.error(f"Error obteniedo el tipo de archivo de {key}. {e}")
    return result


def get_modified_prefix(prefix: str) -> str:
    try:
        if prefix == "/":
            prefix = ""
        elif not prefix.endswith("/") and prefix != "":
            prefix += "/"
    except Exception as e:
        logger.error(f"Error modificando prefijo de {prefix}. {e}")
    finally:
        return prefix


def add_period_to_sufix(name: str, period: str) -> str:
    result = name
    try:
        position = name.rfind(".")
        result = name[:position] + period + name[position:]
    except Exception as e:
        logger.error(f"Error añadiendo periodo al nombre del archivo {name}. {e}")
    return result


def remove_invalid_rows(df: pd.DataFrame, valid_cols: List[str]) -> pd.DataFrame:
    try:
        df = df.dropna(how='all', subset=valid_cols)
    except Exception as e:
        logger.error(f"Error removiendo filas inválidas. {e}")
    finally:
        return df


def remove_fields(df: pd.DataFrame, fields_omitted: List[str]) -> pd.DataFrame:
    try:
        if len(fields_omitted) > 0:
            df = df.loc[:, ~df.columns.isin(fields_omitted)]
    except Exception as e:
        logger.error(f"Error removiendo columnas. {e}")
    finally:
        return df


def update_dict_with_catalogs(data_dict: Dict[str, Any], data: Dict[str, Any], catalog_name: str,
                              default_prefix: str) -> Dict[str, Any]:
    try:
        catalog = data[catalog_name]
        catalog_type = catalog["type"]
        catalog_prefix = catalog["prefix"]
        if catalog_type == CatalogConfigurationEnum.CATALOGO.value:
            catalog_prefix = get_modified_prefix(catalog_prefix)
        else:
            catalog_prefix = default_prefix
        s3_catalog = catalog_prefix + catalog["pattern"]
        data_dict.update({'s3_'+catalog_name: s3_catalog, catalog_name+'_key': catalog["key_field"],
                          catalog_name+'_value': catalog["value_field"]})
        if "delimiter" in catalog.keys():
            data_dict.update({catalog_name+'_delimiter': catalog["delimiter"]})
    except Exception as e:
        raise AssertionError(f"Error actualizando dict de catalogos. {e}")
    finally:
        return data_dict


def update_sql_commands(dataset: List[Tuple[str, str]], label_tablename: str) -> List[Tuple[str, List[str]]]:
    result = []
    allowed_commands = ["create", "update", "delete", "select", "alter", "drop", "begin", "commit"]
    comments = [CommentsScriptEnum[item].value for item in CommentsScriptEnum._member_names_]
    try:
        for row in dataset:
            data = row[1].split(";")
            data = [item.replace("\r", "").replace(";", "").replace("\n", " ") for item in data]
            add_next = False
            final_data = []
            table_name = ""
            for item in data:
                # if item.strip().startswith("--") and label_tablename.strip()+":" not in item:
                #     continue
                if item.lower().strip() == "end":
                    final_data[-1] = final_data[-1] + "; end;"
                final_item = item
                if item.lower().strip().find(label_tablename.lower().strip()+":") != -1:
                    init_index = item.lower().strip().index(label_tablename.lower().strip()+":")
                    table_name = item.replace(" ", "").strip()[init_index+len(label_tablename):].strip()
                    add_next = True
                elif item != "":
                    if add_next:
                        item = table_name + "|" + item
                        add_next = False
                    final_item = item.strip()
                    table_name = ""
                if any([command in final_item.lower() for command in allowed_commands]):
                    if final_item.strip()[:2] in comments:
                        init_indexes = [final_item.lower().find(command) for command in allowed_commands
                                        if final_item.lower().find(command) != -1]
                        if len(init_indexes) > 0:
                            init_index = init_indexes[0]
                            final_item = final_item[init_index:]
                final_item = final_item.replace("%", "%%")
                final_data.append(final_item)
            final_data = [item.replace("\t", "") for item in final_data if item != "" and
                          all([not item.strip().startswith(comment) for comment in comments])]
            result.append((row[0], final_data))
            logger.info(f"Lista de comandos: {result}")
    except Exception as e:
        logger.error(f"Error extrayendo comandos sql. {e}")
    finally:
        return result


def select_multiple(command: str) -> Dict[str, Any]:
    response = {'is_multiple': False, 'tablename': ''}
    tablename = ""
    no_procedure_init = "|select"
    try:
        if command.lower().replace(" ", "").find(no_procedure_init) == -1:
            response["is_multiple"] = True
            tablename = command[:command.index("|")].strip()
            response["tablename"] = tablename
        else:
            if command.lower().replace(" ", "").find(no_procedure_init) != -1:
                response["is_multiple"] = True
                tablename = command[:command.index("|")].strip()
            init_index = command.lower().find("from")
            if init_index == -1:
                raise AssertionError("Query malformed")
            else:
                from_command = command[init_index+4:]
                tablename_base = from_command.strip().split(" ")
                if len(tablename_base) > 0 and tablename == "":
                    tablename = tablename_base[0]
            response["tablename"] = tablename
    except Exception as e:
        raise AssertionError(f"Error validando si es múltiple select y nombre de tabla. {e}")
    finally:
        return response


def create_temp_file(tmp_path: str, filename_mask: str, file_type: str, tablename: str, timezone: str,
                     pattern: str) -> str:
    """ Create an output result as a file with a mask in the name
    """
    fullpath = ""
    try:
        dir_name = str(uuid.uuid4())
        if not tmp_path.endswith("/"):
            tmp_path += "/"
        path_dir = tmp_path + dir_name
        os.mkdir(path_dir)
        current_datetime = str(datetime_by_tzone(timezone, pattern))
        filename_mask = filename_mask.replace("<source_name>", tablename)
        filename_mask = filename_mask.replace("<datetime>", current_datetime)
        filename = filename_mask + "." + file_type
        fullpath = path_dir + "/" + filename
        open(fullpath, mode='a').close()
    except Exception as e:
        logger.error(f"Error creando directorio y archivo temporal.{e}")
    finally:
        return fullpath


def delete_temp_dir(module_name: str) -> bool:
    drop = False
    try:
        if os.path.exists(module_name):
            directory = os.path.dirname(module_name)
            shutil.rmtree(directory, ignore_errors=True)
    except Exception as e:
        raise AssertionError(f"Error borrando modulo temporal. {e}")
    finally:
        return drop


def generateModel(tablename: str, attributes: List[Dict[str, Any]], indexes: List[str], modelName: str = "TableModel"):
    default_precision = 8
    model = type(modelName, (InsumoModel,), {
        '__tablename__': tablename,
        '__table_args__': {'extend_existing': True}
    })
    try:
        for attribute in attributes:
            index = False
            if attribute["name"] in indexes:
                index = True
            logger.debug(f"attribute: {attribute}")
            if attribute["datatype"] == DataTypeEnum.TEXT.name and "maxLength" in attribute.keys():
                setattr(model, attribute["name"],
                        Column(DataTypeOrmEnum[attribute["datatype"]].value(attribute["maxLength"]), index=index))
            elif attribute["datatype"] == DataTypeEnum.DECIMAL.name:
                precision = default_precision
                if "decimal_precision" in attribute.keys():
                    precision = attribute["decimal_precision"]
                setattr(model, attribute["name"],
                        Column(DataTypeOrmEnum[attribute["datatype"]].value(38, precision), index=index))
            else:
                setattr(model, attribute["name"],
                        Column(DataTypeOrmEnum[attribute["datatype"]].value, index=index))
        model = model.__table__
    except InvalidRequestError as e:
        logger.debug(f"InvalidRequestError. {e}")
    except Exception as e:
        logger.error(f"Error creando modelo dinámico. {e}")
    finally:
        return model
